/*
    MonitorIT - A network monitoring tool
    Copyright (C) 2000-2001 Alvaro Lopes <alvieboy@alvie.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __CONFIG_H__
#define __CONFIG_H__

#include "DString.h"
#include "Entity.h"
#include "Visual.h"
#include "config.h"
#include "debug.h"
#include <hash_set>
#include <list>

#include "ConfigItem.h"

class ConfigNode;
class ConfigNodeList;
class ConfigProperty;
class ConfigPropertyList;

class PropertyNotFound {};

class ConfigProperty: public pair<const dstring,dstring>
{
public:
    ConfigProperty(const dstring&, const dstring& v);
    dstring &operator=(const dstring &);
    dstring &operator=(int);
    dstring VALUE() const { return second(); };
    bool operator==(const dstring &);
    const dstring name() const { return first(); }
    const dstring value() const { return second(); }
    //const pair<const dstring,dstring> &Get() const { return *this; }
private:
};

class ConfigPropertyHash
{
public:
    size_t operator() (ConfigProperty const &s) const {
	size_t hash=0x5555;
	const char *p=s.name().c_str();
	for (;*p;p++) hash = (hash<<2) ^ tolower(*p);
	if (hash<0) return -hash;
	return hash;

    }
};
class ConfigPropertyEqual
{
public:
    bool operator() (ConfigProperty const &s,
		    ConfigProperty const &d) const {
        return s.name()==d.name();
    }
};

class ConfigPropertyList: public hash_set<ConfigProperty, ConfigPropertyHash,
ConfigPropertyEqual>
{
public:
    iterator add(const ConfigProperty &c) {
        return insert(c).first;
    }
    ConfigPropertyList& operator+=(const ConfigProperty &p) {
	insert(p);
        return *this;
    }

    dstring Show() const
    {
        dstring res;
        const_iterator it;
	for (it=begin();
	     it!=end();
	     it++) {
	    res+=" ";
	    res+=(*it).name();
	    res+="=\"";
	    res+=(*it).value().Char();
	    res+="\"";
	}
        return res;
    }
};

class ConfigNodeList: public list<ConfigNode*>
{
public:
    ConfigNode *NODE(const dstring &);
};


class ConfigNode
{
public:
    ConfigNode(const dstring &);
    ConfigNode(const dstring &, int un);
    dstring name() { return _name; }
    void NewPROPERTY(const ConfigProperty &);
    void NewPROPERTY_IfNotFound(const ConfigProperty &);
    void NewPROPERTY(const ConfigPropertyList &);
    ConfigProperty *NewPROPERTY(const dstring&);
    void DeletePROPERTY(const dstring&);
    ConfigProperty *PROPERTY(const dstring &n);
    ConfigNode *NODE(const dstring &);
    ConfigNode *NewUniqueNODE(const dstring &n);
  //  ConfigNode *NewUniqueNODE_IfNotFound(const dstring &n);
    ConfigNode *NewNODE(const dstring&);
    void DeleteNODE_IfEmpty(const dstring&);
    ConfigNode *NewNODE_IfNotFound(const dstring&);
    typedef ConfigNodeList::iterator Iterator;
    ConfigNodeList &Childs() { return children; }
    Iterator *FirstLike(const dstring &);
    Iterator *End();
    void setParent(ConfigNode *p) { parent=p; }
    ConfigNode *getParent() { return parent; }
    const ConfigPropertyList prop() const { return properties; }
    int isUnique() { return unique; }
    
private:
    ConfigPropertyList properties;
    ConfigNodeList children;
    ConfigNode *parent;
    
    dstring _name;
    int unique;
};

namespace Gtk
{
    class Box;
    class Label;
};

class ClassTypeHandler
{
public:
    typedef Entity* (*Handler)(ConfigNode*,VisualInterface &) ;
    typedef Gtk::Box* (*GtkConfigurer)();
};
/*
class Configurable
{
public:
    Configurable(const dstring &c, const dstring &v,
		 ClassTypeHandler::Handler h):
    name(c), visualname(v), handler(h)
    {};
    Configurable(const dstring &c, const dstring &v,
		 ClassTypeHandler::Handler h,
		 ClassTypeHandler::GtkConfigurer cg):
    name(c), visualname(v), handler(h),	conf(cg) {};

ClassTypeHandler::GtkConfigurer NP() { return conf; }
ClassTypeHandler::Handler Handler() { return handler; }

const dstring Name() { return name; }
const dstring VisualName() { return visualname; }
protected:
    dstring name, visualname;
    ClassTypeHandler::Handler handler;
    ClassTypeHandler::GtkConfigurer conf;

};
*/



class Configurable: public SigC::Object
{
public:
    virtual Entity *CreateEntity(ConfigNode*, VisualInterface&) = 0;
    virtual const dstring &Name() const = 0;
    virtual const dstring &VisualName() const = 0;

    virtual int Save(ConfigurableItemList &list) = 0;
    virtual ConfigurableItemList *Edit(Entity *e) = 0;
    virtual ConfigurableItemList *Add() = 0;
    virtual ConfigurableItemList *Config() = 0;
    virtual int ApplyEdit( ConfigurableItemList *, Entity *) =0;
    virtual int ApplyConfig( ConfigurableItemList *)=0;
    virtual int ApplyAdd( ConfigurableItemList *)=0;

    virtual void SaveData() = 0;
    virtual void Rehash() = 0;
};

class App;
typedef list<Configurable*> ClassList;

class Config
{
public:
    
    static ConfigNode *NODE(const dstring&);

    //static ClassTypeHandler::Handler GetClassHandler(const dstring &);
    static Entity *Config::CallClassHandler(const dstring &name,
				 ConfigNode *node,
				 VisualInterface &visual);
    static void RegisterClass(Configurable *c);
    typedef ConfigNodeList::iterator Iterator;
    static ConfigNode *NewUniqueNODE(const dstring &n);
    static ConfigNode *NewNODE(const dstring&);
//    static ConfigNode *NewNODE_IfNotFound(const dstring&);

    static ClassList::const_iterator ClassesBegin() {
        return Classes.begin();
    }
    static ClassList::const_iterator ClassesEnd()  {
        return Classes.end();
    }

    static Iterator Begin() {
        return nodes.begin();
    };
    static Iterator End() { return nodes.end(); }
    void Load();
    void NewConfig(const dstring &);
    void LoadFrom(const dstring&);
    static void Save();
    void SaveTo(const dstring&);
    static void DumpTree(FILE*);
    static void DumpTree(FILE*, ConfigNode *c);
    static VisualInterface *GetVisual() { return visual; }
    static void SetVisual(VisualInterface*v) { visual=v; }
    static App *GetApp() { return app; }
    static void SetApp(App*v) { app=v; }

private:
    static ConfigNodeList nodes;
    static list<Configurable*> Classes;
    static VisualInterface *visual;
    static App* app;
    static dstring search;
    static dstring configfile;
};


class ConfigBuilder
{
public:
    ConfigBuilder(Config *c) {
	config = c;
	currentnode = 0;
        start = 1;
    }
public:
    void GetEntity(const dstring&);
    void StartDocument(void);
    void EndDocument(void);
    void StartElement(const dstring &, const dstringhash &);
    void EndElement(const dstring&);
    void Warning(const dstring&);
    void Error(const dstring&);
    void FatalError(const dstring&);
private:
    Config *config;
    ConfigNode *currentnode;
    int start;
};

#ifdef HAVE_GNOME_XML2

#include <libxml/parser.h>

class XMLParser
{
public:
    XMLParser(ConfigBuilder *);
    ~XMLParser();

    void ParseChunk(const dstring &d);
    void ParseFinish();
protected:
    static xmlEntity *xmlGetEntity(void *, const xmlChar*);
    static void xmlStartDocument(void *);
    static void xmlEndDocument(void *);
    static void xmlStartElement(void *, const xmlChar*,
			       const xmlChar **);
    static void xmlEndElement(void *, const xmlChar*);
    static void xmlWarning(void *, const char *format, ...);
    static void xmlError(void *, const char *format, ...);
    static void xmlFatalError(void *, const char *format, ...);
private:
    xmlParserCtxtPtr context;
    static char errmsg[512];
};

#endif


#ifdef HAVE_LIBEXPAT

#include <expat.h>

class XMLParser
{
public:
    XMLParser(ConfigBuilder *);
    ~XMLParser();

    void ParseChunk(const dstring &d);
    void ParseFinish();
protected:
    /*
    static xmlEntity *xmlGetEntity(void *, const xmlChar*);
    static void xmlStartDocument(void *);
    static void xmlEndDocument(void *);
    */
    static void xmlStartElement(void *, const XML_Char*,
			       const XML_Char **);
    static void xmlEndElement(void *, const XML_Char*);
    /*
    static void xmlWarning(void *, const char *format, ...);
    static void xmlError(void *, const char *format, ...);
    static void xmlFatalError(void *, const char *format, ...);
    */
private:
    XML_Parser xmlparser;
    static char errmsg[512];
};

#endif

#endif
